#include "OCC_Connect.h"

#include <vector>
#include <iostream>
#include <iterator>

#include <TopoDS.hxx>
#include <TopoDS_Vertex.hxx>
#include <TopoDS_Solid.hxx>
#include <TopoDS_Compound.hxx>

#include <TopExp.hxx>
#include <TopExp_Explorer.hxx>
#include <TopOpeBRep_Point2d.hxx>
#include <TopOpeBRep_ShapeIntersector.hxx>
#include <TopTools.hxx>
#include <TopTools_IndexedMapOfShape.hxx>

#include <BRepTools.hxx>
#include <BRepTools_ReShape.hxx>
#include <BRepTools_WireExplorer.hxx>
#include <BRepBuilderAPI_MakeVertex.hxx>
#include <BRepBuilderAPI_MakeWire.hxx>
#include <BRepExtrema_ExtCC.hxx>
#include <BRepExtrema_ExtCF.hxx>
#include <BRepExtrema_ExtFF.hxx>
#include <BRepExtrema_ExtPF.hxx>
#include <BRepExtrema_ExtPC.hxx>
#include <BRep_Tool.hxx>
#include <BRepClass_FaceClassifier.hxx>

#include <IntTools_Context.hxx>

#include <BRepExtrema_DistShapeShape.hxx>
#include <BRepOffsetAPI_NormalProjection.hxx>
#include <BRepTopAdaptor_FClass2d.hxx>
#include <BRepTools_Substitution.hxx>
#include <BRepAlgo_Loop.hxx>

#include <BRep_TEdge.hxx>
#include <TopoDS_TShape.hxx>
#include <ShapeFix_Edge.hxx>
#include <Geom_Surface.hxx>


template <typename T>
T max(T const &a, T const &b)
{
    return a<b? b:a;
}

////////////////////////////////////////////////////////////////////////////////
// Printing routines
////////////////////////////////////////////////////////////////////////////////
#include "tprintf.h"

static inline ostream &operator <<(ostream &o,const TopAbs_ShapeEnum &A)
{
    switch(A) {
    case TopAbs_COMPOUND: return o << "COMPOUND";
    case TopAbs_COMPSOLID: return o << "COMPSOLID";
    case TopAbs_SOLID: return o << "SOLID";
    case TopAbs_SHELL: return o << "SHELL";
    case TopAbs_FACE: return o << "FACE";
    case TopAbs_WIRE: return o << "WIRE";
    case TopAbs_EDGE: return o << "EDGE";
    case TopAbs_VERTEX: return o << "VERTEX";
    case TopAbs_SHAPE: return o << "SHAPE";
    default: return o << "Unknown";
    }
}

static inline ostream &operator <<(ostream &o,const TopAbs_State &A)
{
    switch(A) {
    case TopAbs_IN: return o << "IN";
    case TopAbs_OUT: return o << "OUT";
    case TopAbs_ON: return o << "ON";
    case TopAbs_UNKNOWN: return o << "UNKNOWN";
    default: return o << "Unknown";
    }
}

static inline ostream &operator <<(ostream &o,const gp_Pnt &p)
{
    return o << tprintf("(%f %f %f)",p.Coord(1),p.Coord(2),p.Coord(3));
}

static inline ostream &operator <<(ostream &o,const TopOpeBRep_P2Dstatus &A)
{
    switch(A) {
    case TopOpeBRep_P2DUNK: return o << "P2DUNK";
    case TopOpeBRep_P2DINT: return o << "P2DINT";
    case TopOpeBRep_P2DSGF: return o << "P2DSGF";
    case TopOpeBRep_P2DSGL: return o << "P2DSGL ";
    case TopOpeBRep_P2DNEW: return o << "P2DNEW ";
    default: return o << "Unknown";
    }
}

template <typename T>
static inline std::ostream &operator<<(std::ostream &out, std::set<T> const &a)
{ 
    std::copy(a.begin(),a.end(),std::ostream_iterator<T>(out," "));
    return out;
}

bool OCC_Connect::LessThanIntegerSet::operator()(std::set<int> const &a, 
    std::set<int> const &b)
{   
    std::set<int>::iterator pa=a.begin(), pb=b.begin();
    for(; pa!=a.end() && pb!=b.end(); pa++, pb++ ) {
	if(*pa<*pb)
	    return 1;
	if(*pb<*pa)
	    return 0;
    }
    if(pb!=b.end())
	return 1;
    return 0;
}

////////////////////////////////////////////////////////////////////////////////
// OCC_Connect
////////////////////////////////////////////////////////////////////////////////

int OCC_Connect::SaveBRep(char const *name)
{
#if 0
    BRep_Builder BB;
    TopoDS_Compound compound;
    BB.MakeCompound(compound); 
    int i=0;
    for(TopExp_Explorer p(assembly.front(), TopAbs_FACE); p.More(); p.Next()){
        if(++i!=2 && i!=6)
            BB.Add(compound,p.Current());
        else {
            cout << "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB\n";
            BRepTools::Dump(p.Current(),cout);
            cout << "EEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEEE\n";
        }
    }
    BRepTools::Write(compound,name);
#else
    BRepTools::Write(assembly.front(),name);
#endif
}

void OCC_Connect::Dump(ostream &out) const
{
    BRepTools::Dump(assembly.front(),out);
}

void OCC_Connect::Connect(void)
{
    while(assembly.size()>1) {
        TopoDS_Compound result;
        BRep_Builder BB;
        BB.MakeCompound(result);
        Intersect(BB,result,assembly.front(),assembly.back());
        assembly.pop_front();
        assembly.pop_back();
        assembly.push_back(result);
    }
}

void OCC_Connect::Collect(void)
{
    TopoDS_Compound result;
    BRep_Builder BB;
    BB.MakeCompound(result);
    while(assembly.size()>0) {
        if(verbose&Cutting)
	    cout << "Adding item\n";
        BB.Add(result,assembly.front());
        assembly.pop_front();
    }
    assembly.push_back(result);
}

void OCC_Connect::FaceCutters::Add(TopoDS_Wire const &nw)
{
    TopoDS_Vertex nf, nl;
    TopExp::Vertices(nw,nf,nl);
    for(iterator p=begin(); p!=end(); p++) {
	TopoDS_Vertex cf, cl;
	TopoDS_Wire cw=*p;
	TopExp::Vertices(cw,cf,cl);
	if(cf.IsSame(cl))
	    continue;
	if(cf.IsSame(nf) || cf.IsSame(nl) || cl.IsSame(nf) || cl.IsSame(nl)) {
	    BRepBuilderAPI_MakeWire pre(cw);
	    pre.Add(nw);
	    /*  Apparently we need to rebuild to get the edges in order. Some
                algorithms in OCC use an explorer instead of a wire explorer.
             */
	    BRep_Builder BB;
	    TopoDS_Wire rebuilt;
	    BB.MakeWire(rebuilt);
            for(BRepTools_WireExplorer wp(pre); wp.More(); wp.Next()) {
                BB.Add(rebuilt,wp.Current());
            }
            erase(p);
            Add(rebuilt);
	    return;
	}
    }
    push_back(nw);
}

void OCC_Connect::Intersect(BRep_Builder &BB, TopoDS_Shape &target, 
    TopoDS_Shape &shape, TopoDS_Shape &tool)
{
    int t=0;
    /***************************************************************************
      	We start by splitting edges at all the edge-edge intersections.
	This may generate new vertices and edges.
    ***************************************************************************/
    MergeVertices(shape,tool);
    LocOpe_SplitShape splitter1(shape);
    LocOpe_SplitShape splitter2(tool);
    TopOpeBRep_ShapeIntersector intersector;
    for(intersector.InitIntersection(shape,tool);
	intersector.MoreIntersection();
	intersector.NextIntersection()
    ) {
	if(verbose&Cutting) {
	    cout << "++++++++++++++++++++++++++++++++++++++++"
		"++++++++++++++++++++++++++++++++++++++++\n";
	    intersector.DumpCurrent(1); cout << " --> ";
	    intersector.DumpCurrent(2); cout << '\n';
	}

	TopOpeBRep_EdgesIntersector &ee=intersector.ChangeEdgesIntersector();
	if( intersector.CurrentGeomShape(1).ShapeType()==TopAbs_EDGE &&
            intersector.CurrentGeomShape(2).ShapeType()==TopAbs_EDGE
        ) {
	    for(ee.InitPoint(); ee.MorePoint(); ee.NextPoint()) {
		TopOpeBRep_Point2d const &p=ee.Point();
		if(verbose&Cutting)
		    cout << "point loop " << p.Parameter(1) << '\n';
		TopoDS_Vertex vertex;
		if(p.IsVertex(1))
		    vertex=p.Vertex(1);
		else if(p.IsVertex(2))
		    vertex=p.Vertex(2);
		else
		    vertex=BRepBuilderAPI_MakeVertex(p.Value());
		if(!p.IsVertex(1)) {
		    TopoDS_Edge edge=TopoDS::Edge(ee.Edge(1));
		    if(!splitter1.CanSplit(edge)) {
			if(verbose&Cutting)
			    cout << "Cannot split 1\n";;
		    } else {
			if(verbose&Cutting)
			    cout << "splitting model 1\n";
			try { splitter1.Add(vertex,p.Parameter(1),edge); }
			catch(Standard_ConstructionError c) {
			    if(verbose&Cutting)
				cout << "Ooops \n";
			} 
		    }
		}
		if(!p.IsVertex(2)) {
		    TopoDS_Edge edge=TopoDS::Edge(ee.Edge(2));
		    if(!splitter2.CanSplit(edge)) {
			if(verbose&Cutting)
			    cout << "Cannot split 2\n";;
		    } else {
			if(verbose&Cutting)
			    cout << "splitting model 2\n";
			try { splitter2.Add(vertex,p.Parameter(2),edge); }
			catch(Standard_ConstructionError c) {
			    if(verbose&Cutting)
				cout << "Ooops \n";
			}
		    }
		}
	    }
	}
    }

    /***************************************************************************
        Not all intersections seem to be caught, this is an attempt to catch
        some missing intersections. FIXME, this is almost certainly incomplete.
    ***************************************************************************/
    TopTools_IndexedMapOfShape edges, faces, vertices;
    vertices.Clear();
    TopExp::MapShapes(shape,TopAbs_VERTEX,vertices);
    TopExp::MapShapes(tool,TopAbs_VERTEX,vertices);

    edges.Clear();
    TopExp::MapShapes(shape,TopAbs_EDGE,edges);
    for(int e=1; e<=edges.Extent(); e++) {
        TopoDS_Edge edge=TopoDS::Edge(edges(e));
        TopoDS_Vertex o1, o2;
        TopExp::Vertices(edge,o1,o2); 
        int skip1=vertices.FindIndex(o1);
        int skip2=vertices.FindIndex(o2);
        for(int v=1; v<=vertices.Extent(); v++) {
            if(v==skip1 || v==skip2)
                continue;
            TopoDS_Vertex vertex=TopoDS::Vertex(vertices(v));
            BRepExtrema_ExtPC distance(vertex,edge);
            if(!distance.IsDone())
                continue;
            double tolerance=max(BRep_Tool::Tolerance(edge),
                BRep_Tool::Tolerance(vertex));
            for(int i=1;i<=distance.NbExt();i++) {
                if(distance.Value(i)<tolerance) {
                    splitter1.Add(vertex,distance.Parameter(i),edge);
                }
            }
        }
    }

    edges.Clear();
    TopExp::MapShapes(tool,TopAbs_EDGE,edges);
    for(int e=1; e<=edges.Extent(); e++) {
        TopoDS_Edge edge=TopoDS::Edge(edges(e));
        TopoDS_Vertex o1, o2;
        TopExp::Vertices(edge,o1,o2); 
        int skip1=vertices.FindIndex(o1);
        int skip2=vertices.FindIndex(o2);
        for(int v=1; v<=vertices.Extent(); v++) {
            if(v==skip1 || v==skip2)
                continue;
            TopoDS_Vertex vertex=TopoDS::Vertex(vertices(v));
            BRepExtrema_ExtPC distance(vertex,edge);
            if(!distance.IsDone())
                continue;
            double tolerance=max(BRep_Tool::Tolerance(edge),
                BRep_Tool::Tolerance(vertex));
            for(int i=1;i<=distance.NbExt();i++) {
                if(distance.Value(i)<tolerance) {
                    splitter2.Add(vertex,distance.Parameter(i),edge);
                }
            }
        }
    }

    /***************************************************************************
      	We need the shapes with all the edge-edge intersections to split
	all the faces. All vertices and edge which can be merged, will
	be merged.
    ***************************************************************************/
    TopoDS_Compound intermediate1;
    BB.MakeCompound(intermediate1);
    for(TopTools_ListIteratorOfListOfShape p(splitter1.DescendantShapes(shape));
	p.More();
	p.Next()
    ) {
	BB.Add(intermediate1,p.Value());
    }
    TopoDS_Compound intermediate2;
    BB.MakeCompound(intermediate2);
    for(TopTools_ListIteratorOfListOfShape p(splitter2.DescendantShapes(tool));
	p.More();
	p.Next()
    ) {
	BB.Add(intermediate2,p.Value());
    }
    if(verbose&Cutting) {
	cout << "Before merging vertices and edges\n";
	TopoDS_Compound t;
	BB.MakeCompound(t);
	BB.Add(t,intermediate1);
	BB.Add(t,intermediate2);
	PrintItemCount(t);
    }
    MergeVertices(intermediate1,intermediate2);
    MergeEdges(intermediate1,intermediate2);
    if(verbose&Cutting) {
	cout << "After merging vertices and edges\n";
	TopoDS_Compound t;
	BB.MakeCompound(t);
	BB.Add(t,intermediate1);
	BB.Add(t,intermediate2);
	PrintItemCount(t);
    }

    // Create the result
    TopoDS_Compound result;
    BB.MakeCompound(result);
    BB.Add(result,intermediate1);
    BB.Add(result,intermediate2);

    // Add any missing PCurves
    for(TopExp_Explorer face(result,TopAbs_FACE); face.More(); face.Next()) {
	for(TopExp_Explorer edge(face.Current(),TopAbs_EDGE);
	    edge.More();
	    edge.Next()
	) {
            Standard_Real s, e;
            TopoDS_Edge c_edge=TopoDS::Edge(edge.Current());
            TopoDS_Face c_face=TopoDS::Face(face.Current());
            Handle_Geom2d_Curve c=BRep_Tool::CurveOnSurface(c_edge,c_face,s,e);
            if(c.IsNull()) {
                if(verbose&Cutting)
                    cout << "Adding missing PCurve\n";
                ShapeFix_Edge().FixAddPCurve(c_edge,c_face,false,1e-7);
            }
	}
    }

    /***************************************************************************
	We determine which edges/wires are going to cut a face. To do this
	we create a map of FaceCutters which is indexed by the face number
	in the faces map. The FaceCutters generate the correct cutting wires.
    ***************************************************************************/
    edges.Clear();      TopExp::MapShapes(result,TopAbs_EDGE,edges);
    faces.Clear();      TopExp::MapShapes(result,TopAbs_FACE,faces);
    vertices.Clear();   TopExp::MapShapes(result,TopAbs_VERTEX,vertices);
    typedef std::map<int,FaceCutters> cutmap_t;
    cutmap_t cutters;
    for(int face=1;face<=faces.Extent();face++) {
        for(int edge=1;edge<=edges.Extent();edge++) {
            TopoDS_Edge c_edge=TopoDS::Edge(edges(edge));
            TopoDS_Face c_face=TopoDS::Face(faces(face));

            if(BRep_Tool::Degenerated(c_edge)) {
                if(verbose&CuttingReject)
                    cout << "Rejected(" << __LINE__ << ") edge " << edge 
                        << " in face " << face << endl;
                continue;
            }

            TopTools_IndexedMapOfShape already_there;
            TopExp::MapShapes(c_face,TopAbs_EDGE,already_there);
            if(already_there.Contains(c_edge)) {
                if(verbose&CuttingReject)
                    cout << "Rejected(" << __LINE__ << ") edge " << edge 
                        << " in face " << face << endl;
                continue;
            }

            BRepExtrema_DistShapeShape dist(c_edge,c_face);
            double tol=BRep_Tool::Tolerance(c_face); 
            if(dist.Value()>tol) {
                if(verbose&CuttingReject)
                    cout << "Rejected(" << __LINE__ << ") edge " << edge 
                        << " in face " << face << endl;
                continue;
            }

            Standard_Real s, e;
            Handle(Geom_Curve) c3d=BRep_Tool::Curve(c_edge,s,e);
            if(c3d.IsNull()) {
                if(verbose&CuttingReject)
                    cout << "Rejected(" << __LINE__ << ") edge " << edge 
                        << " in face " << face << endl;
                continue;
            }

            // FIXME, checking only the middle may be a bit optimistic.
            TopoDS_Vertex vm=BRepBuilderAPI_MakeVertex(c3d->Value((s+e)/2));
            BRepExtrema_DistShapeShape dist_mf(vm,c_face);
            if(dist_mf.Value()>tol) {
                if(verbose&CuttingReject)
                    cout << "Rejected(" << __LINE__ << ") edge " << edge 
                        << " in face " << face << endl;
                continue;
            }

            if(verbose&Cutting)
		cout << "Accepted edge " << edge << " in face " << face << endl;

            // Add missing PCurve if necessary
            Handle_Geom2d_Curve c=BRep_Tool::CurveOnSurface(c_edge,c_face,s,e);
            if(c.IsNull()) {
                if(verbose&Cutting)
                    cout << "Adding missing PCurve\n";
                ShapeFix_Edge().FixAddPCurve(c_edge,c_face,false,1e-7);
            }
            cutters[face].Add(BRepBuilderAPI_MakeWire(c_edge));
        }
    }

    /***************************************************************************
	Apply all face splits stored in the map.
    ***************************************************************************/
    int cut_count=0;
    LocOpe_SplitShape splitter(result);
    for(cutmap_t::iterator f=cutters.begin(); f!=cutters.end(); f++) {
        TopoDS_Face const &face=TopoDS::Face(faces(f->first));
        FaceCutters &cutter=f->second;
        cut_count+=cutter.size();
	cout << "Cutting face " << f->first << " *************************\n";
	BRepTools::Dump(face,cout);
        for(FaceCutters::iterator p=cutter.begin(); p!=cutter.end(); p++) {
            TopTools_IndexedMapOfShape edges;
	    cout << "With ----------------------------\n";
            TopExp::MapShapes(*p,TopAbs_EDGE,edges);
            if(edges.Extent()<3 && BRep_Tool::IsClosed(*p)) {
                // FIXME This doesn't work.
                cout << "IGNORED Closed wire with less than three edges\n";
                continue;
            } 
	    BRepTools::Dump(*p,cout);
	    splitter.Add(*p,face);
        }
    }
    cout << cut_count << " cuts in " << cutters.size() << " faces\n";

    // Create the final shape with the cutted faces.
    TopoDS_Compound cutted;
    BB.MakeCompound(cutted);
    int count=0;
    for(TopTools_ListIteratorOfListOfShape p(splitter.DescendantShapes(result));
	p.More();
	p.Next()
    ) {
	if(++count==1) {
            cout << "--------- " << count << " ---------------------------\n";
            BB.Add(cutted,p.Value());
            BRepTools::Dump(p.Value(),cout);
        }
    }
    MergeFaces(cutted);
    target=cutted;
}

void OCC_Connect::MergeVertices(TopoDS_Shape &shape1,TopoDS_Shape &shape2) const
{
    TopTools_IndexedMapOfShape imap, omap;
    TopExp::MapShapes(shape1,TopAbs_VERTEX,imap);
    TopExp::MapShapes(shape2,TopAbs_VERTEX,imap);
    BRepTools_ReShape replacer;
    for(int i=0;i<imap.Extent();i++) {
        for(int j=0;j<omap.Extent();j++) {
	    TopoDS_Vertex orig=TopoDS::Vertex(imap(i+1));
	    TopoDS_Vertex repl=TopoDS::Vertex(omap(j+1));
            if(BRepTools::Compare(orig,repl)) {
		repl.Orientation(orig.Orientation());
		replacer.Replace(orig,repl);
		// FIXME, tolerance and point should be updated
                goto skip;
            }
        }
        omap.Add(imap(i+1));
    skip:;
    }
    TopoDS_Shape t=shape1;
    shape1=replacer.Apply(t);
    t=shape2;
    shape2=replacer.Apply(t);
}

bool OCC_Connect::CanMergeCurve(TopoDS_Edge edge1,TopoDS_Edge edge2) const
{
    if(BRep_Tool::Degenerated(edge1) && BRep_Tool::Degenerated(edge2))
	return 1;

    double tol=1e-7;

    Standard_Real s,e;
    Handle(Geom_Curve) curve=BRep_Tool::Curve(edge1,s,e);
    for(Standard_Real i=s;i<=e;i+=(e-s)/10) {
	TopoDS_Vertex v=BRepBuilderAPI_MakeVertex(curve->Value(i));
	double dist=BRepExtrema_DistShapeShape(edge2,v).Value();
	if(dist>tol)
	    return 0;
    }
    return 1;
}

void OCC_Connect::MergeEdges(TopoDS_Shape &shape1, TopoDS_Shape &shape2) const
{
    TopTools_IndexedMapOfShape imap, omap;
    TopExp::MapShapes(shape1,TopAbs_EDGE,imap);
    TopExp::MapShapes(shape2,TopAbs_EDGE,imap);
    BRepTools_ReShape replacer;
    for(int i=0;i<imap.Extent();i++) {
        for(int j=0;j<omap.Extent();j++) {
	    TopoDS_Edge orig=TopoDS::Edge(imap(i+1));
	    TopoDS_Edge repl=TopoDS::Edge(omap(j+1));

	    TopoDS_Vertex o1, o2, r1, r2;
	    TopExp::Vertices(orig,o1,o2,true);
	    TopExp::Vertices(repl,r1,r2,true);

	    if(o1.IsSame(o2)) {
		if(!BRep_Tool::Degenerated(orig)) {
		    if(verbose&Cutting) {
			cout << "Same vertex in edge\n";
			BRepTools::Dump(orig,cout);
		    }
		    replacer.Remove(orig);
		    goto skip;
		} else if(o1.IsSame(r1) && o1.IsSame(r2) 
		    && CanMergeCurve(orig,repl)
		) {
		    if(verbose&Cutting) {
			cout << "Degenerated edge, replace " << i+1 
			    << " with " << j+1 << '\n';
			BRepTools::Dump(orig,cout);
			BRepTools::Dump(repl,cout);
		    }
		    // FIXME, update tolerance
		    BRepTools::Dump(repl.Complemented(),cout);
		    replacer.Replace(orig,repl.Complemented());
		    goto skip;
		}
		cout << i+1 << " Degenerated\n";
	    } 
	    if(o1.IsSame(r1) && o2.IsSame(r2) && CanMergeCurve(orig,repl))  {
                if(verbose&Cutting) {
                    cout << "Same order of vertices, replace " << i+1 
                        << " with " << j+1 << '\n';
                    BRepTools::Dump(orig,cout);
                    BRepTools::Dump(repl,cout);
                }
		// FIXME, update tolerance
		replacer.Replace(orig,repl);
		goto skip;
            }
	    if(o1.IsSame(r2) && o2.IsSame(r1) && CanMergeCurve(orig,repl)) {
                if(verbose&Cutting) {
                    cout << "Reversed order of vertices, replace " << i+1 
                        << " with " << j+1 << '\n';
                    BRepTools::Dump(orig,cout);
                    BRepTools::Dump(repl,cout);
                }
		// FIXME, update tolerance
		replacer.Replace(orig,repl.Complemented());
		goto skip;
            }
        }
        cout << "Adding " << i+1 << " as " << omap.Extent()+1<<" to keep map\n";
        omap.Add(imap(i+1));
    skip:;
    }
    TopoDS_Shape t=shape1;
    shape1=replacer.Apply(t);
    t=shape2;
    shape2=replacer.Apply(t);
}

bool OCC_Connect::CanMergeFace(TopoDS_Face face1,TopoDS_Face face2) const
{
    // FIXME, this really does not work.
    return 1;

    double tol=1e-7;

    Handle(Geom_Surface) surface=BRep_Tool::Surface(face1);
    Standard_Real u1, u2, v1, v2;
    surface->Bounds(u1,u2,v1,v2);
    cout << "u1=" << u1 << " u2=" << u2 << " v1=" << v1 << " v2=" << v2 << endl;
    for(Standard_Real u=u1+(u2-u1)/10;u<u2;u+=(u2-u1)/10) {
        for(Standard_Real v=v1+(v2-v1)/10;v<v2;v+=(v2-v1)/10) {
            TopoDS_Vertex vtx=BRepBuilderAPI_MakeVertex(surface->Value(u,v));
            BRepExtrema_DistShapeShape dss(face2,vtx);
            dss.Perform();
#if 0
            double dist=dss.Value();
            if(dist>tol) {
                cout << "Distance=" << dist << " v=" << v << " u=" << u 
                    << " pnt=" << surface->Value(u,v) << endl;
                BRepTools::Dump(face2,cout);
                return 0;
            }
#else
            dss.Dump(cout);
#endif
        }
    }
    return 1;
}

void OCC_Connect::MergeFaces(TopoDS_Shape &shape) const
{
    /***************************************************************************
        We must find faces which are the same. Since all edges are already
	merged, we know that faces which can be merged have identical edges.
    ***************************************************************************/
    TopTools_IndexedMapOfShape faces, edges;
    TopExp::MapShapes(shape,TopAbs_FACE,faces);
    TopExp::MapShapes(shape,TopAbs_EDGE,edges);
    mapping_t mapping;
    for(int i=0;i<faces.Extent();i++) {
        std::set<int> face_edges;
        for(TopExp_Explorer p(faces(i+1),TopAbs_EDGE); p.More(); p.Next()) {
            int edge=edges.FindIndex(p.Current());
            if(BRep_Tool::Degenerated(TopoDS::Edge(edges(edge)))) {
		cout << "Degenerate edge " << edge << " inserted a 0\n";
		face_edges.insert(0);
	    } else
		face_edges.insert(edge);
        } 
        mapping[face_edges].insert(i+1);
    }

    if(verbose&Cutting) {
        for(mapping_t::iterator p=mapping.begin(); p!=mapping.end(); p++)
            cout << "edges [ " << p->first << "] in face" 
                << (p->second.size()<2? " ":"s ") << "[ " << p->second << "]\n";
    }

    /***************************************************************************
        If two faces have an identical set of edges, they can be merged
	when the planes are never seperated by more than the tolerance. 
    ***************************************************************************/
    BRepTools_ReShape replacer;
    for(mapping_t::iterator p=mapping.begin(); p!=mapping.end(); p++) {
        if(p->second.size()<2)
            continue;
        std::vector<int> uniq;
        for(std::set<int>::iterator q=p->second.begin();
            q!=p->second.end(); 
            q++
        ) {
            for(std::vector<int>::iterator r=uniq.begin(); r!=uniq.end(); r++) {
                TopoDS_Face orig=TopoDS::Face(faces(*q));
                TopoDS_Face repl=TopoDS::Face(faces(*r));
                cout << "Check face " << *q << " and " << *r << endl;
                if(CanMergeFace(orig,repl)) {
                    if(verbose&Cutting) {
                        cout << "replace face " << *q << " with " << *r << '\n';
                    }
                    repl.Orientation(orig.Orientation());
                    replacer.Replace(orig,repl);
                    // FIXME, tolerance should be updated
                    goto skip;
                }
            }
            uniq.push_back(*q);
        skip:;
        }
    }

    TopoDS_Shape t=shape;
    shape=replacer.Apply(t);
}

void OCC_Connect::PrintItemCount(TopoDS_Shape const &shape) const
{
    TopTools_IndexedMapOfShape imap;
    TopExp::MapShapes(shape,TopAbs_VERTEX,imap);
    cout << "Input map contains " << imap.Extent() << " vertices\n";
    imap.Clear();
    TopExp::MapShapes(shape,TopAbs_EDGE,imap);
    cout << "Input map contains " << imap.Extent() << " edges\n";
    imap.Clear();
    TopExp::MapShapes(shape,TopAbs_FACE,imap);
    cout << "Input map contains " << imap.Extent() << " faces\n";
    imap.Clear();
    TopExp::MapShapes(shape,TopAbs_SOLID,imap);
    cout << "Input map contains " << imap.Extent() << " solids\n";
    imap.Clear();
    TopExp::MapShapes(shape,TopAbs_COMPSOLID,imap);
    cout << "Input map contains " << imap.Extent() << " compsolids\n";
    imap.Clear();
    TopExp::MapShapes(shape,TopAbs_COMPOUND,imap);
    cout << "Input map contains " << imap.Extent() << " compounds\n";
}
